@@ -2,7 +2,7 @@
 #
 #   Copyright information
 #
-#	Copyright (C) 2010-2012 Dilshod Temirkhodjaev <tdilshod@gmail.com>
+#	Copyright (C) 2010-2018 Dilshod Temirkhodjaev <tdilshod@gmail.com>
 #
 #   License
 #
@@ -21,9 +21,9 @@
 
 __author__ = "Dilshod Temirkhodjaev <tdilshod@gmail.com>"
 __license__ = "GPL-2+"
-__version__ = "0.6.1"
+__version__ = "0.7.4"
 
-import csv, datetime, zipfile, string, sys, os, re
+import csv, datetime, zipfile, string, sys, os, re, signal
 import xml.parsers.expat
 from xml.dom import minidom
 try:
@@ -76,9 +76,14 @@ FORMATS = {
   'm/d/yyyy' : 'date',
   'dd-mmm-yyyy' : 'date',
   'dd/mm/yyyy' : 'date',
-  'mm/dd/yy hh:mm am/pm' : 'date',
+  'mm/dd/yy h:mm am/pm' : 'date',
+  'mm/dd/yy hh:mm' : 'date',
+  'mm/dd/yyyy h:mm am/pm' : 'date',
   'mm/dd/yyyy hh:mm:ss' : 'date',
   'yyyy-mm-dd hh:mm:ss' : 'date',
+  '#,##0;(#,##0)' : 'float',
+  '_(* #,##0_);_(* (#,##0);_(* "-"??_);_(@_)' : 'float',
+  '_(* #,##0.00_);_(* (#,##0.00);_(* "-"??_);_(@_)' : 'float'
 }
 STANDARD_FORMATS = {
   0 : 'general',
@@ -126,36 +131,46 @@ class OutFileAlreadyExistsException(XlsxException):
 class Xlsx2csv:
     """
      Usage: Xlsx2csv("test.xslx", **params).convert("test.csv", sheetid=1)
+     Input:
+       xlsxfile - path to file or filehandle
      options:
        sheetid - sheet no to convert (0 for all sheets)
        dateformat - override date/time format
+       timeformat - override time format
+       floatformat - override float format
+       quoting - if and how to quote
        delimiter - csv columns delimiter symbol
        sheetdelimiter - sheets delimiter used when processing all sheets
        skip_empty_lines - skip empty lines
+       skip_trailing_columns - skip trailing columns
        hyperlinks - include hyperlinks
        include_sheet_pattern - only include sheets named matching given pattern
        exclude_sheet_pattern - exclude sheets named matching given pattern
     """
 
     def __init__(self, xlsxfile, **options):
-        # dateformat=None, delimiter=",", sheetdelimiter="--------", skip_empty_lines=False, escape_strings=False, cmd=False
-        options.setdefault("delimeter", ",")
+        options.setdefault("delimiter", ",")
+        options.setdefault("quoting", csv.QUOTE_MINIMAL)
         options.setdefault("sheetdelimiter", "--------")
+        options.setdefault("dateformat", None)
+        options.setdefault("timeformat", None)
+        options.setdefault("floatformat", None)
+        options.setdefault("scifloat", False)
         options.setdefault("skip_empty_lines", False)
+        options.setdefault("skip_trailing_columns", False)
         options.setdefault("escape_strings", False)
-        options.setdefault("cmd", False)
+        options.setdefault("hyperlinks", False)
         options.setdefault("include_sheet_pattern", ["^.*$"])
         options.setdefault("exclude_sheet_pattern", [])
         options.setdefault("merge_cells", False)
+        options.setdefault("ignore_formats", [''])
+        options.setdefault("lineterminator", "\n")
 
         self.options = options
         try:
             self.ziphandle = zipfile.ZipFile(xlsxfile)
         except (zipfile.BadZipfile, IOError):
-            if self.options['cmd']:
-                sys.stderr.write("Invalid xlsx file: " + xlsxfile + os.linesep)
-                sys.exit(1)
-            raise InvalidXlsxFileException("Invalid xlsx file: " + xlsxfile)
+            raise InvalidXlsxFileException("Invalid xlsx file: " + str(xlsxfile))
 
         self.py3 = sys.version_info[0] == 3
 
@@ -166,6 +181,10 @@ class Xlsx2csv:
         if self.options['escape_strings']:
             self.shared_strings.escape_strings()
 
+    def __del__(self):
+        # make sure to close zip file, ziphandler does have a close() method
+        self.ziphandle.close()
+
     def getSheetIdByName(self, name):
         for s in self.workbook.sheets:
             if s['name'] == name:
@@ -181,10 +200,7 @@ class Xlsx2csv:
                 if not os.path.exists(outfile):
                     os.makedirs(outfile)
                 elif os.path.isfile(outfile):
-                    if self.options['cmd']:
-                        sys.stderr.write("File " + outfile + " already exists!" + os.linesep)
-                        sys.exit(1)
-                    raise OutFileAlreadyExistsException("File " + outfile + " already exists!")
+                    raise OutFileAlreadyExistsException("File " + str(outfile) + " already exists!")
             for s in self.workbook.sheets:
                 sheetname = s['name']
 
@@ -219,34 +235,49 @@ class Xlsx2csv:
                 if isinstance(outfile, str):
                     of = os.path.join(outfile, sheetname + '.csv')
                 elif self.options['sheetdelimiter'] and len(self.options['sheetdelimiter']):
-                    of.write(self.options['sheetdelimiter'] + " " + str(s['id']) + " - " + sheetname + os.linesep)
+                    of.write(self.options['sheetdelimiter'] + " " + str(s['id']) + " - " + sheetname + self.options['lineterminator'])
                 self._convert(s['id'], of)
 
     def _convert(self, sheetid, outfile):
         closefile = False
         if isinstance(outfile, str):
-            outfile = open(outfile, 'w+')
+            if sys.version_info[0] == 2:
+                outfile = open(outfile, 'wb+')
+            elif sys.version_info[0] == 3:
+                outfile = open(outfile, 'w+', encoding=self.options['outputencoding'], newline="")
+            else:
+                sys.stderr.write("error: version of your python is not supported: " + str(sys.version_info) + "\n")
+                sys.exit(1)
             closefile = True
         try:
-            writer = csv.writer(outfile, quoting=csv.QUOTE_MINIMAL, delimiter=self.options['delimiter'], lineterminator=os.linesep)
+            writer = csv.writer(outfile, quoting=self.options['quoting'], delimiter=self.options['delimiter'], lineterminator=self.options['lineterminator'])
             sheetfile = self._filehandle("xl/worksheets/sheet%i.xml" % sheetid)
+            if not sheetfile:
+                sheetfile = self._filehandle("xl/worksheets/worksheet%i.xml" % sheetid)
             if not sheetfile and sheetid == 1:
                 sheetfile = self._filehandle("xl/worksheets/sheet.xml")
+            if not sheetfile and sheetid == 1:
+                sheetfile = self._filehandle("xl/worksheets/worksheet.xml")
             if not sheetfile:
-                if self.options['cmd']:
-                    sys.stderr.write("Sheet %s not found!%s" %(sheetid, os.linesep))
-                    sys.exit(1)
                 raise SheetNotFoundException("Sheet %s not found" %sheetid)
+            sheet = Sheet(self.workbook, self.shared_strings, self.styles, sheetfile)
             try:
-                sheet = Sheet(self.workbook, self.shared_strings, self.styles, sheetfile)
                 sheet.relationships = self._parse(Relationships, "xl/worksheets/_rels/sheet%i.xml.rels" % sheetid)
                 sheet.set_dateformat(self.options['dateformat'])
+                sheet.set_timeformat(self.options['timeformat'])
+                sheet.set_floatformat(self.options['floatformat'])
                 sheet.set_skip_empty_lines(self.options['skip_empty_lines'])
+                sheet.set_skip_trailing_columns(self.options['skip_trailing_columns'])
                 sheet.set_include_hyperlinks(self.options['hyperlinks'])
                 sheet.set_merge_cells(self.options['merge_cells'])
+                sheet.set_scifloat(self.options['scifloat'])
+                sheet.set_ignore_formats(self.options['ignore_formats'])
+                if self.options['escape_strings'] and sheet.filedata:
+                    sheet.filedata = re.sub(r"(<v>[^<>]+)&#10;([^<>]+</v>)", r"\1\\n\2", re.sub(r"(<v>[^<>]+)&#9;([^<>]+</v>)", r"\1\\t\2", re.sub(r"(<v>[^<>]+)&#13;([^<>]+</v>)", r"\1\\r\2", sheet.filedata)))
                 sheet.to_csv(writer)
             finally:
                 sheetfile.close()
+                sheet.close()
         finally:
             if closefile:
                 outfile.close()
@@ -308,7 +339,7 @@ class Workbook:
         for sheetNode in sheetNodes:
             attrs = sheetNode._attrs
             name = attrs["name"].value
-            if self.appName == 'xl':
+            if self.appName == 'xl' and len(attrs["r:id"].value) > 2:
                 if 'r:id' in attrs: id = int(attrs["r:id"].value[3:])
                 else: id = int(attrs['sheetId'].value)
             else:
@@ -340,7 +371,7 @@ class Relationships:
                 target = attrs.get('Target')
                 self.relationships[str(rId.value)] = {
                     "type" : vtype and str(vtype.value) or None,
-                    "target" : target and str(target.value) or None
+                    "target" : target and target.value.encode("utf-8") or None
                 }
 
 class Styles:
@@ -401,6 +432,11 @@ class SharedStrings:
             self.value+= data
 
     def handleStartElement(self, name, attrs):
+        # ignore namespace
+        i = name.find(":")
+        if i >= 0:
+            name = name[i + 1:]
+
         if name == 'si':
             self.si = True
             self.value = ""
@@ -412,6 +448,11 @@ class SharedStrings:
             self.rPh = True
 
     def handleEndElement(self, name):
+        # ignore namespace
+        i = name.find(":")
+        if i >= 0:
+            name = name[i + 1:]
+
         if name == 'si':
             self.si = False
             self.strings.append(self.value)
@@ -434,17 +475,21 @@ class Sheet:
         self.in_row = False
         self.in_cell = False
         self.in_cell_value = False
-        self.in_cell_formula = False
 
         self.columns = {}
+        self.lastRowNum = 0
         self.rowNum = None
         self.colType = None
         self.cellId = None
         self.s_attr = None
         self.data = None
+        self.max_columns = -1
 
         self.dateformat = None
+        self.timeformat = "%H:%M" # default time format
+        self.floatformat = None
         self.skip_empty_lines = False
+        self.skip_trailing_columns = False
 
         self.filedata = None
         self.filehandle = filehandle
@@ -454,13 +499,31 @@ class Sheet:
 
         self.hyperlinks = {}
         self.mergeCells = {}
+        self.ignore_formats = []
+
+    def close(self):
+        # Make sure Worksheet is closed, parsers lib does not have a close() function, so simply delete it
+        self.parser = None
 
     def set_dateformat(self, dateformat):
         self.dateformat = dateformat
 
+    def set_timeformat(self, timeformat):
+        if timeformat:
+            self.timeformat = timeformat
+
+    def set_floatformat(self, floatformat):
+        self.floatformat = floatformat
+
     def set_skip_empty_lines(self, skip):
         self.skip_empty_lines = skip
 
+    def set_skip_trailing_columns(self, skip):
+        self.skip_trailing_columns = skip
+
+    def set_ignore_formats(self, ignore_formats):
+        self.ignore_formats = ignore_formats
+
     def set_merge_cells(self, mergecells):
         if not mergecells:
             return
@@ -499,7 +562,10 @@ class Sheet:
                     for cell in self._range(rangeStr):
                         self.mergeCells[cell] = {}
                         self.mergeCells[cell]['copyFrom'] = rng[0]
-        
+
+    def set_scifloat(self, scifloat):
+        self.scifloat = scifloat
+
     def set_include_hyperlinks(self, hyperlinks):
         if not hyperlinks or not self.relationships or not self.relationships.relationships:
             return
@@ -549,6 +615,7 @@ class Sheet:
     def to_csv(self, writer):
         self.writer = writer
         self.parser = xml.parsers.expat.ParserCreate()
+        self.parser.buffer_text = True
         self.parser.CharacterDataHandler = self.handleCharData
         self.parser.StartElementHandler = self.handleStartElement
         self.parser.EndElementHandler = self.handleEndElement
@@ -565,31 +632,36 @@ class Sheet:
                 self.data = self.sharedStrings[int(self.data)]
             elif self.colType == "b": # boolean
                 self.data = (int(data) == 1 and "TRUE") or (int(data) == 0 and "FALSE") or data
+            elif self.colType == "str" or self.colType == "inlineStr":
+                self.data = data
             elif self.s_attr:
                 s = int(self.s_attr)
 
                 # get cell format
-                format = None
+                format_str = None
                 xfs_numfmt = self.styles.cellXfs[s]
                 if xfs_numfmt in self.styles.numFmts:
-                    format = self.styles.numFmts[xfs_numfmt]
+                    format_str = self.styles.numFmts[xfs_numfmt]
                 elif xfs_numfmt in STANDARD_FORMATS:
-                    format = STANDARD_FORMATS[xfs_numfmt]
+                    format_str = STANDARD_FORMATS[xfs_numfmt]
                 # get format type
-                if not format:
+                if not format_str:
+                    print("unknown format %s at %d" %(format_str,xfs_numfmt))
                     return
                 format_type = None
-                if format in FORMATS:
-                    format_type = FORMATS[format]
-                elif re.match("^\d+(\.\d+)?$", self.data) and re.match(".*[hsmdyY]", format) and not re.match('.*\[.*[dmhys].*\]', format):
+                if format_str in FORMATS:
+                    format_type = FORMATS[format_str]
+                elif re.match("^\d+(\.\d+)?$", self.data) and re.match(".*[hsmdyY]", format_str) and not re.match('.*\[.*[dmhys].*\]', format_str):
                     # it must be date format
                     if float(self.data) < 1:
                         format_type = "time"
                     else:
                         format_type = "date"
-                elif re.match("^\d", self.data):
+                elif re.match("^-?\d+(.\d+)?$", self.data) or (self.scifloat and re.match("^-?\d+(.\d+)?([eE]-?\d+)?$", self.data)):
                     format_type = "float"
-                if format_type:
+                if format_type == 'date' and self.dateformat == 'float':
+                    format_type = "float"
+                if format_type and not format_type in self.ignore_formats:
                     try:
                         if format_type == 'date': # date/time
                             if self.workbook.date1904:
@@ -601,27 +673,39 @@ class Sheet:
                                 self.data = date.strftime(str(self.dateformat))
                             else:
                                 # ignore ";@", don't know what does it mean right now
-                                dateformat = format.replace(";@", ""). \
+                                # ignore "[$-409]" and similar format codes
+                                dateformat = re.sub(r"\[\$\-\d{1,3}\]", "", format_str, 1). \
+                                  replace(";@", ""). \
                                   replace("yyyy", "%Y").replace("yy", "%y"). \
-                                  replace("hh:mm", "%H:%M").replace("h", "%H").replace("%H%H", "%H").replace("ss", "%S"). \
-                                  replace("d", "%e").replace("%e%e", "%d"). \
-                                  replace("mmmm", "%B").replace("mmm", "%b").replace(":mm", ":%M").replace("m", "%m").replace("%m%m", "%m"). \
-                                  replace("am/pm", "%p")
+                                  replace("hh:mm", "%H:%M").replace("h", "%I").replace("%H%H", "%H").replace("ss", "%S"). \
+                                  replace("dd", "d").replace("d", "%d"). \
+                                  replace("am/pm", "%p"). \
+                                  replace("mmmm", "%B").replace("mmm", "%b").replace(":mm", ":%M").replace("m", "%m").replace("%m%m", "%m")
                                 self.data = date.strftime(str(dateformat)).strip()
                         elif format_type == 'time': # time
-                            t = int(float(self.data) * 24*60)
-                            self.data = "%.2i:%.2i" %(t / 60, t % 60)  #str(t / 60) + ":" + ('0' + str(t % 60))[-2:]
+                            t = int(round((float(self.data) % 1) * 24*60*60, 6)) # it should be in seconds
+                            d = datetime.time((t / 3600) % 24, (t / 60) % 60, t % 60)
+                            self.data = d.strftime(self.timeformat)
                         elif format_type == 'float' and ('E' in self.data or 'e' in self.data):
+                            self.data = str(self.floatformat or '%f') % float(self.data)
+                        # if cell is general, be aggressive about stripping any trailing 0s, decimal points, etc.
+                        elif format_type == 'float' and format_str == 'general':
                             self.data = ("%f" %(float(self.data))).rstrip('0').rstrip('.')
-                        elif format_type == 'float' and format[0:3] == '0.0':
-                            self.data = ("%." + str(len(format.split(".")[1])) + "f") % float(self.data)
-                            
-                    except (ValueError, OverflowError):
+                        elif format_type == 'float' and format_str[0:3] == '0.0':
+                            if self.floatformat:
+                                self.data = str(self.floatformat) % float(self.data)
+                            else:
+                                L = len(format_str.split(".")[1])
+                                if '%' in format_str:
+                                    L += 1
+                                self.data = ("%." + str(L) + "f") % float(self.data)
+                        elif format_type == 'float':
+                            # unsupported float formatting
+                            self.data = ("%f" %(float(self.data))).rstrip('0').rstrip('.')
+
+                    except (ValueError, OverflowError): # this catch must be removed, it's hiding potential problems
                         # invalid date format
                         pass
-        # does not support it
-        #elif self.in_cell_formula:
-        #    self.formula = data
 
     def handleStartElement(self, name, attrs):
         has_namespace = name.find(":") > 0
@@ -634,21 +718,21 @@ class Sheet:
                 self.colIndex = 0
             else:
                 self.colIndex+= 1
-            #self.formula = None
             self.data = ""
             self.in_cell = True
         elif self.in_cell and ((name == 'v' or name == 'is') or (has_namespace and (name.endswith(':v') or name.endswith(':is')))):
             self.in_cell_value = True
             self.collected_string = ""
-        #elif self.in_cell and name == 'f':
-        #    self.in_cell_formula = True
         elif self.in_sheet and (name == 'row' or (has_namespace and name.endswith(':row'))) and ('r' in attrs):
             self.rowNum = attrs['r']
             self.in_row = True
             self.columns = {}
             self.spans = None
             if 'spans' in attrs:
-                self.spans = [int(i) for i in attrs['spans'].split(":")]
+                self.spans = [int(i) for i in attrs['spans'].split(" ")[-1].split(":")]
+        elif name == 't':
+            # reset collected string
+            self.collected_string = ""
 
         elif name == 'sheetData' or (has_namespace and name.endswith(':sheetData')):
             self.in_sheet = True
@@ -656,21 +740,18 @@ class Sheet:
             rng = attrs.get("ref").split(":")
             if len(rng) > 1:
                 start = re.match("^([A-Z]+)(\d+)$", rng[0])
-                end = re.match("^([A-Z]+)(\d+)$", rng[1])
-                startCol = start.group(1)
-                #startRow = int(start.group(2))
-                endCol = end.group(1)
-                #endRow = int(end.group(2))
-                self.columns_count = 0
-                for cell in self._range(startCol + "1:" + endCol + "1"):
-                    self.columns_count+= 1
+                if (start):
+                    end = re.match("^([A-Z]+)(\d+)$", rng[1])
+                    startCol = start.group(1)
+                    endCol = end.group(1)
+                    self.columns_count = 0
+                    for cell in self._range(startCol + "1:" + endCol + "1"):
+                        self.columns_count+= 1
 
     def handleEndElement(self, name):
         has_namespace = name.find(":") > 0
-        if self.in_cell and name == 'v':
+        if self.in_cell and ((name == 'v' or name == 'is' or name == 't') or (has_namespace and (name.endswith(':v') or name.endswith(':is')))):
             self.in_cell_value = False
-        #elif self.in_cell and name == 'f':
-        #    self.in_cell_formula = False
         elif self.in_cell and (name == 'c' or (has_namespace and name.endswith(':c'))):
             t = 0
             for i in self.colNum: t = t*26 + ord(i) - 64
@@ -678,11 +759,12 @@ class Sheet:
             if self.hyperlinks:
                 hyperlink = self.hyperlinks.get(self.cellId)
                 if hyperlink:
+                    hyperlink = hyperlink.decode("utf-8")
                     d = "<a href='" + hyperlink + "'>" + d + "</a>"
             if self.colNum + self.rowNum in self.mergeCells.keys():
                 if 'copyFrom' in self.mergeCells[self.colNum + self.rowNum].keys() and self.mergeCells[self.colNum + self.rowNum]['copyFrom'] == self.colNum + self.rowNum:
                     self.mergeCells[self.colNum + self.rowNum]['value'] = d
-                elif len(d) == len(''):
+                else:
                     d = self.mergeCells[self.mergeCells[self.colNum + self.rowNum]['copyFrom']]['value']
 
             self.columns[t - 1 + self.colIndex] = d
@@ -696,14 +778,31 @@ class Sheet:
                         val = val.encode("utf-8")
                     d[k] = val
                 if self.spans:
-                    l = self.spans[0] + self.spans[1] - 1
+                    l = self.spans[1]
                     if len(d) < l:
                         d+= (l - len(d)) * ['']
+
+                # write empty lines
+                if not self.skip_empty_lines:
+                    for i in range(self.lastRowNum, int(self.rowNum) - 1):
+                        self.writer.writerow([])
+                    self.lastRowNum = int(self.rowNum)
+
                 # write line to csv
                 if not self.skip_empty_lines or d.count('') != len(d):
                     while len(d) < self.columns_count:
                         d.append("")
+
+                    if self.skip_trailing_columns:
+                        if self.max_columns < 0:
+                            self.max_columns = len(d)
+                            while len(d) > 0 and d[-1] == "":
+                                d = d[0:-1]
+                                self.max_columns = self.max_columns - 1
+                        elif self.max_columns > 0:
+                            d = d[0:self.max_columns]
                     self.writer.writerow(d)
+
             self.in_row = False
         elif self.in_sheet and (name == 'sheetData' or (has_namespace and name.endswith(':sheetData'))):
             self.in_sheet = False
@@ -738,16 +837,12 @@ class Sheet:
 
 
 def convert_recursive(path, sheetid, outfile, kwargs):
-    kwargs['cmd'] = False
     for name in os.listdir(path):
         fullpath = os.path.join(path, name)
         if os.path.isdir(fullpath):
             convert_recursive(fullpath, sheetid, outfile, kwargs)
         else:
-            # strange code, python2.4 fix
-            #outfilepath = outfile if len(outfile) > 0 else ""
             outfilepath = outfile
-
             if len(outfilepath) == 0 and fullpath.lower().endswith(".xlsx"):
                 outfilepath = fullpath[:-4] + 'csv'
 
@@ -758,11 +853,17 @@ def convert_recursive(path, sheetid, outfile, kwargs):
                 print("File %s is not a zip file" %fullpath)
 
 if __name__ == "__main__":
+    try:
+        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
+        signal.signal(signal.SIGINT, signal.SIG_DFL)
+    except AttributeError:
+        pass
+
     if "ArgumentParser" in globals():
         parser = ArgumentParser(description = "xlsx to csv converter")
         parser.add_argument('infile', metavar='xlsxfile', help="xlsx file path")
         parser.add_argument('outfile', metavar='outfile', nargs='?', help="output csv file path")
-        parser.add_argument('-v', '--version', action='version', version='%(prog)s')
+        parser.add_argument('-v', '--version', action='version', version=__version__)
         nargs_plus = "+"
         argparser = True
     else:
@@ -771,30 +872,51 @@ if __name__ == "__main__":
         nargs_plus = 1
         argparser = False
 
+
+    if sys.version_info[0] == 2 and sys.version_info[1] < 5:
+        inttype = "int"
+    else:
+        inttype = int
     parser.add_argument("-a", "--all", dest="all", default=False, action="store_true",
       help="export all sheets")
-    parser.add_argument("-s", "--sheet", dest="sheetid", default=1, type=int,
-      help="sheet number to convert")
-    parser.add_argument("-n", "--sheetname", dest="sheetname", default=None,
-      help="sheet name to convert")
+    parser.add_argument("-c", "--outputencoding", dest="outputencoding", default="utf-8", action="store",
+      help="encoding of output csv ** Python 3 only ** (default: utf-8)")
     parser.add_argument("-d", "--delimiter", dest="delimiter", default=",",
       help="delimiter - columns delimiter in csv, 'tab' or 'x09' for a tab (default: comma ',')")
-    parser.add_argument("-f", "--dateformat", dest="dateformat",
-      help="override date/time format (ex. %%Y/%%m/%%d)")
-    parser.add_argument("-i", "--ignoreempty", dest="skip_empty_lines", default=False, action="store_true",
-      help="skip empty lines")
-    parser.add_argument("-e", "--escape", dest='escape_strings', default=False, action="store_true",
-      help="Escape \\r\\n\\t characters")
-    parser.add_argument("-p", "--sheetdelimiter", dest="sheetdelimiter", default="--------",
-      help="sheet delimiter used to separate sheets, pass '' if you do not need delimiter (default: '--------')")
     parser.add_argument("--hyperlinks", "--hyperlinks", dest="hyperlinks", action="store_true", default=False,
       help="include hyperlinks")
-    parser.add_argument("-I", "--include_sheet_pattern", nargs=nargs_plus, dest="include_sheet_pattern", default="^.*$",
-      help="only include sheets named matching given pattern, only effects when -a option is enabled.")
+    parser.add_argument("-e", "--escape", dest='escape_strings', default=False, action="store_true",
+      help="Escape \\r\\n\\t characters")
     parser.add_argument("-E", "--exclude_sheet_pattern", nargs=nargs_plus, dest="exclude_sheet_pattern", default="",
       help="exclude sheets named matching given pattern, only effects when -a option is enabled.")
+    parser.add_argument("-f", "--dateformat", dest="dateformat",
+      help="override date/time format (ex. %%Y/%%m/%%d)")
+    parser.add_argument("-t", "--timeformat", dest="timeformat",
+      help="override time format (ex. %%H/%%M/%%S)")
+    parser.add_argument("--floatformat", dest="floatformat",
+      help="override float format (ex. %%.15f)")
+    parser.add_argument("--sci-float", dest="scifloat", default=False, action="store_true",
+      help="force scientific notation to float")
+    parser.add_argument("-I", "--include_sheet_pattern", nargs=nargs_plus, dest="include_sheet_pattern", default="^.*$",
+      help="only include sheets named matching given pattern, only effects when -a option is enabled.")
+    parser.add_argument("--ignore-formats", nargs=nargs_plus, type=str, dest="ignore_formats", default=[''],
+      help="Ignores format for specific data types.")
+    parser.add_argument("-l", "--lineterminator", dest="lineterminator", default="\n",
+      help="line terminator - lines terminator in csv, '\\n' '\\r\\n' or '\\r' (default: \\n)")
     parser.add_argument("-m", "--merge-cells", dest="merge_cells", default=False, action="store_true",
       help="merge cells")
+    parser.add_argument("-n", "--sheetname", dest="sheetname", default=None,
+      help="sheet name to convert")
+    parser.add_argument("-i", "--ignoreempty", dest="skip_empty_lines", default=False, action="store_true",
+      help="skip empty lines")
+    parser.add_argument("--skipemptycolumns", dest="skip_trailing_columns", default=False, action="store_true",
+      help="skip trailing empty columns")
+    parser.add_argument("-p", "--sheetdelimiter", dest="sheetdelimiter", default="--------",
+      help="sheet delimiter used to separate sheets, pass '' if you do not need delimiter, or 'x07' or '\\f' for form feed (default: '--------')")
+    parser.add_argument("-q", "--quoting", dest="quoting", default="minimal",
+      help="quoting - fields quoting in csv, 'none' 'minimal' 'nonnumeric' or 'all' (default: minimal)")
+    parser.add_argument("-s", "--sheet", dest="sheetid", default=1, type=inttype,
+      help="sheet number to convert")
 
     if argparser:
         options = parser.parse_args()
@@ -808,40 +930,88 @@ if __name__ == "__main__":
         options.outfile = len(args) > 1 and args[1] or None
 
     if len(options.delimiter) == 1:
-        delimiter = options.delimiter
-    elif options.delimiter == 'tab':
-        delimiter = '\t'
+        pass
+    elif options.delimiter == 'tab' or options.delimiter == '\\t':
+        options.delimiter = '\t'
     elif options.delimiter == 'comma':
-        delimiter = ','
+        options.delimiter = ','
     elif options.delimiter[0] == 'x':
-        delimiter = chr(int(options.delimiter[1:]))
+        options.delimiter = chr(int(options.delimiter[1:]))
     else:
-        raise XlsxException("Invalid delimiter")
+        sys.stderr.write("error: invalid delimiter\n")
+        sys.exit(1)
+
+    if options.quoting == 'none':
+        options.quoting = csv.QUOTE_NONE
+    elif options.quoting == 'minimal':
+        options.quoting = csv.QUOTE_MINIMAL
+    elif options.quoting == 'nonnumeric':
+        options.quoting = csv.QUOTE_NONNUMERIC
+    elif options.quoting == 'all':
+        options.quoting = csv.QUOTE_ALL
+    else:
+        sys.stderr.write("error: invalid quoting\n")
+        sys.exit(1)
+
+    if options.lineterminator == '\n':
+        pass
+    elif options.lineterminator == '\\n':
+        options.lineterminator = '\n'
+    elif options.lineterminator == '\\r':
+        options.lineterminator = '\r'
+    elif options.lineterminator == '\\r\\n':
+        options.lineterminator = '\r\n'
+    else:
+        sys.stderr.write("error: invalid line terminator\n")
+        sys.exit(1)
+
+    if options.sheetdelimiter == '--------':
+        pass
+    elif options.sheetdelimiter == '':
+        pass
+    elif options.sheetdelimiter == '\\f':
+        options.sheetdelimiter = '\f'
+    elif options.sheetdelimiter[0] == 'x':
+        options.sheetdelimiter = chr(int(options.sheetdelimiter[1:]))
+    else:
+        sys.stderr.write("error: invalid sheet delimiter\n")
+        sys.exit(1)
 
     kwargs = {
-      'delimiter' : delimiter,
+      'delimiter' : options.delimiter,
+      'quoting' : options.quoting,
       'sheetdelimiter' : options.sheetdelimiter,
       'dateformat' : options.dateformat,
+      'timeformat' : options.timeformat,
+      'floatformat' : options.floatformat,
+      'scifloat' : options.scifloat,
       'skip_empty_lines' : options.skip_empty_lines,
+      'skip_trailing_columns' : options.skip_trailing_columns,
       'escape_strings' : options.escape_strings,
       'hyperlinks' : options.hyperlinks,
-      'cmd' : True,
       'include_sheet_pattern' : options.include_sheet_pattern,
       'exclude_sheet_pattern' : options.exclude_sheet_pattern,
-      'merge_cells' : options.merge_cells
+      'merge_cells' : options.merge_cells,
+      'outputencoding' : options.outputencoding,
+      'lineterminator' : options.lineterminator,
+      'ignore_formats' : options.ignore_formats
     }
     sheetid = options.sheetid
     if options.all:
         sheetid = 0
 
     outfile = options.outfile or sys.stdout
-    if os.path.isdir(options.infile):
-        convert_recursive(options.infile, sheetid, outfile, kwargs)
-    else:
-        xlsx2csv = Xlsx2csv(options.infile, **kwargs)
-        if options.sheetname:
-            sheetid = xlsx2csv.getSheetIdByName(options.sheetname)
-            if not sheetid:
-                raise XlsxException("Sheet '%s' not found" % options.sheetname)
-
-        xlsx2csv.convert(outfile, sheetid)
+    try:
+        if os.path.isdir(options.infile):
+            convert_recursive(options.infile, sheetid, outfile, kwargs)
+        else:
+            xlsx2csv = Xlsx2csv(options.infile, **kwargs)
+            if options.sheetname:
+                sheetid = xlsx2csv.getSheetIdByName(options.sheetname)
+                if not sheetid:
+                    raise XlsxException("Sheet '%s' not found" % options.sheetname)
+            xlsx2csv.convert(outfile, sheetid)
+    except XlsxException:
+        _, e, _ = sys.exc_info()
+        sys.stderr.write(str(e) + "\n")
+        sys.exit(1)
